/*
 * Copyright (C) 2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file adu.h Global definitions. */

#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h> /* time(), localtime() */
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <inttypes.h>
#include <string.h>
#include <assert.h>
#include <osl.h>
#include "gcc-compat.h"
#include "select.cmdline.h"

/** debug loglevel, gets really noisy */
#define DEBUG 1
/** still noisy, but won't fill your disk */
#define INFO  2
/** normal, but significant event */
#define NOTICE 3
/** unexpected event that can be handled */
#define WARNING 4
/** unhandled error condition */
#define ERROR 5
/** system might be unreliable */
#define CRIT 6
/** last message before exit */
#define EMERG 7

/** Log messages with lower priority than that will not be compiled in. */
#define COMPILE_TIME_LOGLEVEL 0

/** \cond */
#if DEBUG > COMPILE_TIME_LOGLEVEL
#define DEBUG_LOG(f,...) __log(DEBUG, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define DEBUG_LOG(...) do {;} while (0)
#endif

#if INFO > COMPILE_TIME_LOGLEVEL
#define INFO_LOG(f,...) __log(INFO, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define INFO_LOG(...) do {;} while (0)
#endif

#if NOTICE > COMPILE_TIME_LOGLEVEL
#define NOTICE_LOG(f,...) __log(NOTICE, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define NOTICE_LOG(...) do {;} while (0)
#endif

#if WARNING > COMPILE_TIME_LOGLEVEL
#define WARNING_LOG(f,...) __log(WARNING, "%s: " f, __FUNCTION__, ##  __VA_ARGS__)
#else
#define WARNING_LOG(...) do {;} while (0)
#endif

#if ERROR > COMPILE_TIME_LOGLEVEL
#define ERROR_LOG(f,...) __log(ERROR, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define ERROR_LOG(...) do {;} while (0)
#endif

#if CRIT > COMPILE_TIME_LOGLEVEL
#define CRIT_LOG(f,...) __log(CRIT, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define CRIT_LOG(...) do {;} while (0)
#endif

#if EMERG > COMPILE_TIME_LOGLEVEL
#define EMERG_LOG(f,...) __log(EMERG, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define EMERG_LOG(...)
#endif
/** \endcond */

/**
 * Write a log message to a dynamically allocated string.
 *
 * \param fmt Usual format string.
 * \param p Result pointer.
 *
 * \sa printf(3). */
#define VSPRINTF(fmt, p) \
{ \
	int n; \
	size_t size = 100; \
	p = adu_malloc(size); \
	while (1) { \
		va_list ap; \
		/* Try to print in the allocated space. */ \
		va_start(ap, fmt); \
		n = vsnprintf(p, size, fmt, ap); \
		va_end(ap); \
		/* If that worked, return the string. */ \
		if (n > -1 && n < size) \
			break; \
		/* Else try again with more space. */ \
		if (n > -1) /* glibc 2.1 */ \
			size = n + 1; /* precisely what is needed */ \
		else /* glibc 2.0 */ \
			size *= 2; /* twice the old size */ \
		p = adu_realloc(p, size); \
	} \
}

/** Evaluates to 1 if x < y, to -1 if x > y and to 0 if x == y. */
#define NUM_COMPARE(x, y) ((int)((x) < (y)) - (int)((x) > (y)))

/** The columns of the directory table. */
enum dir_table_columns {
	/** The name of the directory. */
	DT_NAME,
	/** The dir count number. */
	DT_NUM,
	/** The number of the parent directory. */
	DT_PARENT_NUM,
	/** The number of bytes of all regular files. */
	DT_BYTES,
	/** The number of all regular files. */
	DT_FILES,
	/** Number of columns in this table. */
	NUM_DT_COLUMNS
};

/** The columns of the id table. */
enum user_table_columns {
	/** The numer of the directory. */
	UT_DIR_NUM,
	/** The number of bytes of all regular files in this dir owned by this id. */
	UT_BYTES,
	/** The number of files in this dir owned by this id. */
	UT_FILES,
	/** Number of columns in this table. */
	NUM_UT_COLUMNS
};

/** Flags for the user hash table. */
enum uid_info_flags {
	/** Whether this slot of the hash table is used. */
	UI_FL_SLOT_USED = 1,
	/** Whether this uid should be taken into account. */
	UI_FL_ADMISSIBLE = 2,
};

/** Information about one admissible user. */
struct user_info {
	/** User ID. */
	uint32_t uid;
	/** \sa enum uid_info_flags. */
	uint32_t flags;
	/** The user name.  */
	char *pw_name;
	/** The user table of this user.*/
	struct osl_table *table;
	/** Total number of files owned by this user. */
	uint64_t files;
	/** Total number of bytes owned by this user. */
	uint64_t bytes;
	/** Total number of directories that contain at least one file */
	uint64_t dirs;
	/** The description of the user table. */
	struct osl_table_description *desc;
};

/**
 * Describes one range of admissible user IDs.
 *
 * adu converts the admissible user ids given at the command line
 * into an array of such structs.
 */
struct uid_range {
	/** Lowest admissible user ID. */
	uint32_t low;
	/** Greatest admissible user ID. */
	uint32_t high;
};

enum search_uid_flags {
	OPEN_USER_TABLE = 1,
	CREATE_USER_TABLE = 2,
};

#define FOR_EACH_UID_RANGE(ur, urs) for (ur = urs; ur->low <= ur->high; ur++)

extern uint32_t num_uids;
extern struct osl_table *dir_table;

/** The adu command line options. */
extern struct gengetopt_args_info conf;

/**
 * The select command line options.
 *
 * Either given at the command line, or via the \a set command
 * in interactive mode.
 */
extern struct select_args_info select_conf;

/* adu.c */
__printf_2_3 void __log(int, const char*, ...);
int open_dir_table(int create);
void check_signals(void);
void close_all_tables(void);
char *get_uid_list_name(void);
void create_hash_table(unsigned bits);
int search_uid(uint32_t uid, struct uid_range *urs,
		enum search_uid_flags flags, struct user_info **ui_ptr);
int for_each_admissible_user(int (*func)(struct user_info *, void *),
		void *data);
void sort_hash_table(int (*comp)(const void *, const void *));

/* create.c */
int com_create(void);

/* interactive.c */
void print_interactive_help(void);
int com_interactive(void);
