#include <ctype.h> /* isspace() */

#include "adu.h"
#include "format.h"
#include "select.h"
#include "string.h"
#include "error.h"
#include "cmdline.h"

/**
 * Describes one valid command for interactive mode.
 *
 * When invoked in interactive mode, adu reads commands from stdin. There's a
 * static array of all such commands.
 */
struct interactive_command {
	/** The name of the command. */
	const char *name;
	/** Pointer to The function that is being executed. */
	int (*handler)(char *);
	/** Help text. */
	const char *desc;
};

static struct uid_range *admissible_uids;
static struct format_info *fi;

#define INTERACTIVE_COMMANDS \
	INTERACTIVE_COMMAND(set, "change the current configuration") \
	INTERACTIVE_COMMAND(reset, "reset configuration to defaults") \
	INTERACTIVE_COMMAND(help, "show list of commands and one-line descriptions") \
	INTERACTIVE_COMMAND(run, "start the query according to the current configuration")


#define INTERACTIVE_COMMAND(name, desc) \
	static int icom_ ## name (char *line);

INTERACTIVE_COMMANDS

#undef INTERACTIVE_COMMAND

#define INTERACTIVE_COMMAND(_name, _desc) \
	{ \
	.name = #_name, \
	.handler = icom_ ## _name, \
	.desc = _desc \
	},

struct interactive_command icmds[] = {
	INTERACTIVE_COMMANDS
	{.name  = NULL}
};

#define FOR_EACH_COMMAND(c) for (c = icmds; c->name; c++)

static int read_input_line(char *line, size_t size)
{
	return fgets(line, size, stdin)? 1 : -1;
}

static int icom_run(__a_unused char *line)
{
	return run_select_query(admissible_uids, fi);
}

static int icom_help(__a_unused char *line)
{
	struct interactive_command *c;

	FOR_EACH_COMMAND(c)
		fprintf(stdout, "%s\t%s\n", c->name, c->desc);
	return 1;
}

void print_interactive_help(void)
{
	struct interactive_command *c;
	FOR_EACH_COMMAND(c)
		fprintf(stdout, "\t%s\t%s\n", c->name, c->desc);
}

static int icom_reset(__a_unused char *line)
{
	free_format_info(fi);
	fi = NULL;
	free(admissible_uids);
	admissible_uids = NULL;
	select_cmdline_parser_init(&select_conf);
	return 1;
}

static int icom_set(char *line)
{
	struct select_cmdline_parser_params params = {
		.override = 1,
		.initialize = 0,
		.check_required = 1,
		.check_ambiguity = 0,
		.print_errors = 1
	};
	if (!line) {
		select_cmdline_parser_dump(stdout, &select_conf);
		return 1;
	}

	free_format_info(fi);
	fi = NULL;
	free(admissible_uids);
	admissible_uids = NULL;
	return parse_select_options(line, &params, &admissible_uids, &fi);
}

static int exec_interactive_command(char *line)
{
	const char const *delim = "\t\n\f\r\v ";
	int i;
	char *cmd, *args;
	int ret = -E_SYNTAX;
	size_t len;

	if (!line || !*line)
		return 1;
	len = strlen(line);

	while (len && isspace(line[len - 1])) {
		line[len - 1] = '\0';
		len--;
	}
	if (!len)
		return 1;
	line += strspn(line, delim); /* skip initial whitespace */
	if (!*line)
		return 1;
	/* OK, we have a non-empty line */
	if (*line == '#')
		return 1;
	cmd = adu_strdup(line);
	args = cmd + strcspn(cmd, delim);
	if (!*args)
		args = NULL;
	else {
		*args = '\0';
		args++;
		/* let p point to the next non-whitespace char */
		args += strspn(args, delim);
		if (!*args)
			args = NULL;
	}
	DEBUG_LOG("name: %s, args: %s.\n", cmd, args);
	for (i = 0; icmds[i].name; i++) {
		if (strcmp(icmds[i].name, cmd))
			continue;
		INFO_LOG("exec cmd: %s, args: %s\n", cmd, args);
		ret = icmds[i].handler(args);
		break;
	}
	free(cmd);
	return ret;
}

int com_interactive(void)
{
	char line[255];
	int ret = 1;

	select_cmdline_parser_init(&select_conf);
	ret = parse_select_options(NULL, NULL, &admissible_uids, &fi);
	while (read_input_line(line, sizeof(line)) >= 0) {
		ret = exec_interactive_command(line);
		if (ret < 0)
			printf("%s\n", adu_strerror(-ret));
	}
	return ret;
}
