/*
 * Copyright (C) 2004-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file string.c Memory allocation and string handling functions. */

#include "adu.h"
#include "string.h"
#include <string.h>
#include "error.h"

/**
 * Paraslash's version of realloc().
 *
 * \param p Pointer to the memory block, may be \p NULL.
 * \param size The desired new size.
 *
 * A wrapper for realloc(3). It calls \p exit(\p EXIT_FAILURE) on errors,
 * i.e. there is no need to check the return value in the caller.
 *
 * \return A pointer to  the newly allocated memory, which is suitably aligned
 * for any kind of variable and may be different from \a p.
 *
 * \sa realloc(3).
 */
__must_check __malloc void *adu_realloc(void *p, size_t size)
{
	/*
	 * No need to check for NULL pointers: If p is NULL, the  call
	 * to realloc is equivalent to malloc(size)
	 */
	assert(size);
	if (!(p = realloc(p, size))) {
		EMERG_LOG("realloc failed (size = %zu), aborting\n",
			size);
		exit(EXIT_FAILURE);
	}
	return p;
}

/**
 * Paraslash's version of malloc().
 *
 * \param size The desired new size.
 *
 * A wrapper for malloc(3) which exits on errors.
 *
 * \return A pointer to the allocated memory, which is suitably aligned for any
 * kind of variable.
 *
 * \sa malloc(3).
 */
__must_check __malloc void *adu_malloc(size_t size)
{
	assert(size);
	void *p = malloc(size);

	if (!p) {
		EMERG_LOG("malloc failed (size = %zu),  aborting\n",
			size);
		exit(EXIT_FAILURE);
	}
	return p;
}

/**
 * Paraslash's version of calloc().
 *
 * \param size The desired new size.
 *
 * A wrapper for calloc(3) which exits on errors.
 *
 * \return A pointer to the allocated and zeroed-out memory, which is suitably
 * aligned for any kind of variable.
 *
 * \sa calloc(3)
 */
__must_check __malloc void *adu_calloc(size_t size)
{
	void *ret = adu_malloc(size);

	memset(ret, 0, size);
	return ret;
}

/**
 * Paraslash's version of strdup().
 *
 * \param s The string to be duplicated.
 *
 * A wrapper for strdup(3). It calls \p exit(EXIT_FAILURE) on errors, i.e.
 * there is no need to check the return value in the caller.
 *
 * \return A pointer to the duplicated string. If \p s was the NULL pointer,
 * an pointer to an empty string is returned.
 *
 * \sa strdup(3)
 */
__must_check __malloc char *adu_strdup(const char *s)
{
	char *ret;

	if ((ret = strdup(s? s: "")))
		return ret;
	EMERG_LOG("strdup failed, aborting\n");
	exit(EXIT_FAILURE);
}

/**
 * Allocate a sufficiently large string and print into it.
 *
 * \param fmt A usual format string.
 *
 * Produce output according to \p fmt. No artificial bound on the length of the
 * resulting string is imposed.
 *
 * \return This function either returns a pointer to a string that must be
 * freed by the caller or aborts without returning.
 *
 * \sa printf(3).
 */
__must_check __printf_1_2 __malloc char *make_message(const char *fmt, ...)
{
	char *msg;

	VSPRINTF(fmt, msg);
	return msg;
}

/**
 * adu's version of strcat().
 *
 * \param a String to be appended to.
 * \param b String to append.
 *
 * Append \p b to \p a.
 *
 * \return If \a a is \p NULL, return a pointer to a copy of \a b, i.e.
 * para_strcat(NULL, b) is equivalent to para_strdup(b). If \a b is \p NULL,
 * return \a a without making a copy of \a a.  Otherwise, construct the
 * concatenation \a c, free \a a (but not \a b) and return \a c.
 *
 * \sa strcat(3).
 */
__must_check __malloc char *adu_strcat(char *a, const char *b)
{
	char *tmp;

	if (!a)
		return adu_strdup(b);
	if (!b)
		return a;
	tmp = make_message("%s%s", a, b);
	free(a);
	return tmp;
}

/** \cond LLONG_MAX and LLONG_LIN might not be defined. */
#ifndef LLONG_MAX
#define LLONG_MAX (1 << (sizeof(long) - 1))
#endif
#ifndef LLONG_MIN
#define LLONG_MIN (-LLONG_MAX - 1LL)
#endif
/** \endcond */

/**
 * Convert a string to a 64-bit signed integer value.
 *
 * \param str The string to be converted.
 * \param result Result pointer.
 *
 * \return Standard.
 *
 * \sa strtol(3), atoi(3).
 */
__must_check int atoi64(const char *str, int64_t *result)
{
	char *endptr;
	long long tmp;

	errno = 0; /* To distinguish success/failure after call */
	tmp = strtoll(str, &endptr, 10);
	if (errno == ERANGE && (tmp == LLONG_MAX || tmp == LLONG_MIN))
		return -E_ATOI_OVERFLOW;
	if (errno != 0 && tmp == 0) /* other error */
		return -E_STRTOLL;
	if (endptr == str)
		return -E_ATOI_NO_DIGITS;
	if (*endptr != '\0') /* Further characters after number */
		return -E_ATOI_JUNK_AT_END;
	*result = tmp;
	return 1;
}

/**
 * Split string and return pointers to its parts.
 *
 * \param args The string to be split.
 * \param argv_ptr Pointer to the list of substrings.
 * \param delim Delimiter.
 *
 * This function modifies \a args by replacing each occurance of \a delim by
 * zero. A \p NULL-terminated array of pointers to char* is allocated dynamically
 * and these pointers are initialized to point to the broken-up substrings
 * within \a args. A pointer to this array is returned via \a argv_ptr.
 *
 * \return The number of substrings found in \a args.
 */
__must_check unsigned split_args(char *args, char *** const argv_ptr, const char *delim)
{
	char *p = args;
	char **argv;
	size_t n = 0, i, j;

	p = args + strspn(args, delim);
	for (;;) {
		i = strcspn(p, delim);
		if (!i)
			break;
		p += i;
		n++;
		p += strspn(p, delim);
	}
	*argv_ptr = adu_malloc((n + 1) * sizeof(char *));
	argv = *argv_ptr;
	i = 0;
	p = args + strspn(args, delim);
	while (p) {
		argv[i] = p;
		j = strcspn(p, delim);
		if (!j)
			break;
		p += strcspn(p, delim);
		if (*p) {
			*p = '\0';
			p++;
			p += strspn(p, delim);
		}
		i++;
	}
	argv[n] = NULL;
	return n;
}

static int check_uid_arg(const char *arg, uint32_t *uid)
{
	const uint32_t max = ~0U;
	/*
	 * we need an 64-bit int for string -> uid conversion because strtoll()
	 * returns a signed value.
	 */
	int64_t val;
	int ret = atoi64(arg, &val);

	if (ret < 0)
		return ret;
	if (val < 0 || val > max)
		return -ERRNO_TO_ERROR(EINVAL);
	*uid = val;
	return 1;
}

int parse_uid_range(const char *orig_arg, struct uid_range *ur)
{
	int ret;
	char *arg = adu_strdup(orig_arg), *p = strchr(arg, '-');

	if (!p || p == arg) { /* -42 or 42 */
		ret = check_uid_arg(p? p + 1 : arg, &ur->high);
		if (ret < 0)
			goto out;
		ur->low = p? 0 : ur->high;
		ret = 1;
		goto out;
	}
	/* 42- or 42-4711 */
	*p = '\0';
	p++;
	ret = check_uid_arg(arg, &ur->low);
	if (ret < 0)
		goto out;
	ur->high = ~0U;
	if (*p) { /* 42-4711 */
		ret = check_uid_arg(p, &ur->high);
		if (ret < 0)
			goto out;
	}
	if (ur->low > ur->high)
		ret = -ERRNO_TO_ERROR(EINVAL);
out:
	if (ret < 0)
		ERROR_LOG("bad uid option: %s\n", orig_arg);
	else
		INFO_LOG("admissible uid range: %u - %u\n", ur->low,
			ur->high);
	free(arg);
	return ret;
}

int parse_uid_arg(const char *orig_arg, struct uid_range **ur)
{
	char *arg, **argv;
	unsigned n;
	int i, ret = 1;

	if (!orig_arg)
		return 0;
	arg = adu_strdup(orig_arg);
	n = split_args(arg, &argv, ",");
	if (!n)
		return -E_SYNTAX;
	*ur = adu_malloc((n + 1) * sizeof(struct uid_range));
	for (i = 0; i < n; i++) {
		ret = parse_uid_range(argv[i], *ur + i);
		if (ret < 0)
			break;
	}
	free(argv);
	free(arg);
	if (ret < 0) {
		free(*ur);
		*ur = NULL;
		return ret;
	}
	/* an empty range indicates the end of the list */
	(*ur)[n].low = 1;
	(*ur)[n].high = 0;
	return n;
}

enum line_state_flags {LSF_HAVE_WORD = 1, LSF_BACKSLASH = 2, LSF_QUOTE = 4};

static int get_next_word(const char *line, char **word)
{
	const char *in;
	char *out;
	int ret, state = 0;

	out = adu_malloc(strlen(line) + 1);
	*out = '\0';
	*word = out;
	for (in = line; *in; in++) {
		switch (*in) {
		case '\\':
			if (state & LSF_BACKSLASH) /* \\ */
				break;
			state |= LSF_BACKSLASH;
			state |= LSF_HAVE_WORD;
			continue;
		case 'n':
		case 't':
			if (state & LSF_BACKSLASH) { /* \n or \t */
				*out++ = (*in == 'n')? '\n' : '\t';
				state &= ~LSF_BACKSLASH;
				continue;
			}
			break;
		case '"':
			if (state & LSF_BACKSLASH) /* \" */
				break;
			if (state & LSF_QUOTE) {
				state &= ~LSF_QUOTE;
				continue;
			}
			state |= LSF_HAVE_WORD;
			state |= LSF_QUOTE;
			continue;
		case ' ':
		case '\t':
		case '\n':
			if (state & LSF_BACKSLASH)
				break;
			if (state & LSF_QUOTE)
				break;
			if (state & LSF_HAVE_WORD)
				goto success;
			/* ignore space at the beginning */
			continue;
		}
		/* copy char */
		state |= LSF_HAVE_WORD;
		*out++ = *in;
		state &= ~LSF_BACKSLASH;
	}
	ret = 0;
	if (!(state & LSF_HAVE_WORD))
		goto out;
	ret = -ERRNO_TO_ERROR(EINVAL);
	if (state & LSF_BACKSLASH) {
		ERROR_LOG("trailing backslash\n");
		goto out;
	}
	if (state & LSF_QUOTE) {
		ERROR_LOG("unmatched quote character\n");
		goto out;
	}
success:
	*out = '\0';
	return in - line;
out:
	free(*word);
	*word = NULL;
	return ret;
}

void free_argv(char **argv)
{
	int i;

	for (i = 0; argv[i]; i++)
		free(argv[i]);
	free(argv);
}

/**
 * \return Number of words in \a line, negative on errors.
 */
int create_argv(const char *line, char ***result)
{
	char *word, **argv = adu_malloc(2 * sizeof(char *));
	const char *p;
	int ret, num_words;

	argv[0] = adu_strdup(line);
	for (p = line, num_words = 1; ; p += ret, num_words++) {
		ret = get_next_word(p, &word);
		if (ret < 0)
			goto err;
		if (!ret)
			break;
		argv = adu_realloc(argv, (num_words + 2) * sizeof(char*));
		argv[num_words] = word;
	}
	argv[num_words] = NULL;
	*result = argv;
	return num_words;
err:
	while (num_words > 0)
		free(argv[--num_words]);
	free(argv);
	return ret;
}
